"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.dataset.common import (
    ProcessType, SubProcessType,
    Topology, Physics,
    Track
)
from arrakis_nd.arrakis.common import (
    tracklette_data_type,
    track_data_type
)


class DeltaPlugin(Plugin):
    """
    This plugin labels hits which are associated to ionized electrons
    generated from MIPs and HIPS (i.e. delta electrons).

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(DeltaPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = ['tracklette', 'track', 'blip']

        if "delta_size_threshold" not in self.config:
            self.config["delta_size_threshold"] = 10
        self.delta_size_threshold = self.config["delta_size_threshold"]

        self.delta_labels = {
            'topology': Topology.Track.value,
            'physics': Physics.DeltaElectron.value
        }
        self.low_energy_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.ElectronIonization.value,
        }
        self.mip_labels = {
            'topology': Topology.Track.value,
            'physics': Physics.MIP.value,
        }
        self.hip_labels = {
            'topology': Topology.Track.value,
            'physics': Physics.HIP.value,
        }

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']
        parent_pdg_ids = event_products['parent_pdg_id']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_E = trajectories['E_start']
        trajectories_start_process = trajectories['start_process']
        trajectories_start_subprocess = trajectories['start_subprocess']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the mips/hips which result in ionization"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 11) &
            (
                (abs(parent_pdg_ids) == 13) |
                (abs(parent_pdg_ids) == 15) |
                (abs(parent_pdg_ids) == 211) |
                (abs(parent_pdg_ids) == 321) |
                (abs(parent_pdg_ids) == 2212)
            ) &
            (
                (abs(trajectories_start_process) == ProcessType.Electromagnetic.value) &
                (abs(trajectories_start_subprocess) == SubProcessType.Ionization.value)
            )
        )

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if mip has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            particle_segments = track_id_hit_segment_map[(particle_id, vertex_id)]
            particle_hit_segments = (
                particle_hits, particle_segments
            )

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """############################### Set standard labels ###############################"""
            """Set the event_id"""
            arrakis_charge['event_id'][particle_hits] = event

            """############################### Delta Electrons ###############################"""
            """
            Delta electrons are electrons which have are ionized from MIPs/HIPs, and
            have enough energy to leave behind some visible track signature.  Technically
            speaking, all ionization electrons are deltas, but most of them leave no visible
            energy deposits in the detector.  Thus, we must define some parameter(s) that
            we use to control what we consider to be deltas in truth.

            Deltas are daughter electrons of MIPs/HIPs which should have been created
            from electromagnetic ionization processes, which is the criteria used for
            filtering them.
            """
            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[particle_mask][ii]
            particle_xyz_end = trajectories_xyz_end[particle_mask][ii]
            particle_pxyz_start = trajectories_pxyz_start[particle_mask][ii]
            particle_pxyz_end = trajectories_pxyz_end[particle_mask][ii]
            particle_E = trajectories_E[particle_mask][ii]
            """Determine track begin and end points"""
            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(particle_start_distances)
            closest_end_index = np.argmin(particle_end_distances)

            particle_start_index = (
                particle_hits[closest_start_index],
                particle_segments[closest_start_index]
            )
            particle_end_index = (
                particle_hits[closest_end_index],
                particle_segments[closest_end_index]
            )

            """Add delta/michel vertex"""
            if (
                (len(particle_hits) >= self.delta_size_threshold) &
                (closest_start_index != closest_end_index)
            ):
                arrakis_charge['vertex'][particle_start_index] = 1
                arrakis_charge['tracklette_end'][particle_end_index] = 1

                """Iterate over delta labels"""
                for label, value in self.delta_labels.items():
                    arrakis_charge[label][particle_hit_segments] = value
            else:
                if abs(parent_pdg_ids[particle_mask][ii]) == 2212:
                    """Iterate over hip labels"""
                    for label, value in self.hip_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
                else:
                    """Iterate over mip labels"""
                    for label, value in self.mip_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value

            """Set the particle label"""
            arrakis_charge['particle'][particle_hit_segments] = trajectories_pdg_ids[particle_mask][ii]

            """Generate unique label"""
            arrakis_charge['unique_topology'][particle_hit_segments] = traj_index

            """Parameterize the trajectory of this track using t0"""
            track_data = fit_track(particle_hit_t0s, particle_charge_xyz)

            track_type = Track.Delta.value
            """############################### Construct Tracklettes ###############################"""

            """
            Now we must determine if the track is broken along its
            trajectory.  We can do this by checking if the beginning and
            ending points are in different TPCs.  If so, then we will have
            to determine the beginning and ending points of each tracklette.

            We can do this by isolating hits according to the io_group,
            which defines the particular TPC region that a hit is in.  For
            each unique io_group associated to the hits of this particle,
            we will follow the same procedure and find the closest
            points to the track beginning and ending to get the tracklette
            begin and end points.
            """
            particle_io_group = charge_io_group[particle_hits]
            unique_particle_io_group = np.unique(particle_io_group)
            particle_hits = np.array(particle_hits)
            particle_segments = np.array(particle_segments)
            particle_hit_t0s = np.array(particle_hit_t0s)
            for io_group in unique_particle_io_group:
                """Isolate hits from this io_group"""
                io_group_mask = (particle_io_group == io_group)
                if not sum(io_group_mask):
                    continue
                io_group_hits = particle_hits[io_group_mask]
                io_group_segments = particle_segments[io_group_mask]
                io_group_t0s = particle_hit_t0s[io_group_mask]

                """Find the closest points to track begin/end for this io_group"""
                io_group_start_distances = particle_start_distances[io_group_mask]
                io_group_end_distances = particle_end_distances[io_group_mask]
                tracklette_start_index = np.argmin(io_group_start_distances)
                tracklette_end_index = np.argmin(io_group_end_distances)

                io_group_start_index = (
                    io_group_hits[tracklette_start_index],
                    io_group_segments[tracklette_start_index]
                )
                io_group_end_index = (
                    io_group_hits[tracklette_end_index],
                    io_group_segments[tracklette_end_index]
                )

                if tracklette_start_index == tracklette_end_index:
                    continue

                arrakis_charge['tracklette_begin'][io_group_start_index] = 1
                arrakis_charge['tracklette_end'][io_group_end_index] = 1

                """Parameterize the trajectory of this tracklette using t0"""
                io_group_xyz = particle_charge_xyz[io_group_mask]
                track_data = fit_track(io_group_t0s, io_group_xyz)

                """Now generate the associated tracklette CAF object"""
                """
                The tracklette data object has the following entries
                that must be filled (see arrakis.common).
                """
                io_group_xyz = particle_charge_xyz[io_group_mask]

                """Create the CAF object for this tracklette"""
                tracklette_data = np.array([(
                    event,
                    io_group,
                    particle_id,
                    vertex_id,
                    track_type,
                    [0, 0, 0],
                    [0, 0, 0],
                    [io_group_xyz[tracklette_start_index]],
                    [io_group_xyz[tracklette_end_index]],
                    [0, 0, 0],
                    [0, 0, 0],
                    [track_data['track_dir']],
                    [track_data['track_enddir']],
                    sum(particle_charge_E[io_group_mask]),
                    1,
                    track_data['track_len_gcm2'],
                    track_data['track_len_cm'],
                    0,
                    [[traj_index]+[0]*19],
                    [[1]+[0]*19])],
                    dtype=tracklette_data_type
                )

                """Add the new tracklette to the CAF objects"""
                event_products['tracklette'].append(tracklette_data)

            """############################### Construct Track ###############################"""
            """Now generate the associated track CAF object"""
            """
            The track data object has the following entries
            that must be filled (see arrakis.common).
            """
            io_group_xyz = particle_charge_xyz[io_group_mask]

            """Create the CAF object for this track"""
            particle_pxyz_start_magnitude = np.linalg.norm(particle_pxyz_start)
            particle_pxyz_end_magnitude = np.linalg.norm(particle_pxyz_end)
            if particle_pxyz_start_magnitude > 0:
                particle_pxyz_start /= particle_pxyz_start_magnitude
            if particle_pxyz_end_magnitude > 0:
                particle_pxyz_end /= particle_pxyz_end_magnitude

            track_data = np.array([(
                event,
                particle_id,
                vertex_id,
                track_type,
                particle_xyz_start,
                particle_xyz_end,
                [particle_charge_xyz[closest_start_index]],
                [particle_charge_xyz[closest_end_index]],
                particle_pxyz_start,
                particle_pxyz_end,
                [track_data['track_dir']],
                [track_data['track_enddir']],
                sum(particle_charge_E),
                1,
                track_data['track_len_gcm2'],
                track_data['track_len_cm'],
                particle_E,
                [[traj_index]+[0]*19],
                [[1]+[0]*19])],
                dtype=track_data_type
            )

            """Add the new track to the CAF objects"""
            event_products['track'].append(track_data)

        """Write changes to arrakis_file"""
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



import plotly
import plotly.graph_objects as go


def custom_legend_name(fig,new_names):
    '''
    Custom legend name for plotly figures

    Args:
        fig    (px.Figure): plotly figure
        new_names (list):      list of new names for the legend

    Returns:
        fig (px.Figure): plotly figure with the custom legend names
    '''
    for i, new_name in enumerate(new_names):
        fig.data[i].name = new_name
        # fig.data[i].hovertemplate = new_name
    # fig.for_each_trace(lambda t: t.update(name = dict[t.name],legendgroup = dict[t.name],hovertemplate = t.hovertemplate.replace(t.name, dict[t.name])))
    return fig

def custom_plotly_layout(fig,xaxis_title ="",yaxis_title="",title="",legend_pos=["out","h"],log_menu=True,fontsize=16,figsize=None,ranges=(None,None),matches=("x","y"),
    tickformat=('.s','.s'),barmode="stack",margin={"auto":True}):
    '''
    Custom layout for plotly figures

    Args:
        fig      (Figure):       plotly figure
        xaxis_title (str):       x axis title (default: "")
        yaxis_title (str):       y axis title (default: "")
        title       (str):       title (default: "")
        legend_pos  (list):      legend position (default: ["out","h"])
        log_menu    (bool):      log menu (default: True)
        fontsize    (int):       font size (default: 16)
        figsize     (tuple):     figure size (default: None)
        ranges      (tuple):     axis ranges (default: (None,None))
        matches     (tuple):     axis matches (default: ("x","y"))
        tickformat  (tuple):     axis tick format (default: ('.s','.s'))
        barmode     (str):       barmode (default: "stack")
        margin      (dict):      figure margin (default: {"auto":True,"color":"white","margin":(0,0,0,0)})
    
    Returns:
        fig (Figure): plotly figure with the custom layout
    '''
    default_margin = {"color":"white","margin":(0,0,0,0)}
    if margin != None:
            for key in default_margin.keys():
                if key not in margin.keys():
                    margin[key] = default_margin[key]

    legend_opt = {"in": dict(yanchor="top", xanchor="right", x=0.99),"out": dict(yanchor="bottom", xanchor="right", x=1.2,y=0.2) }
    if log_menu: fig.update_layout( updatemenus=[ dict( buttons=list([ dict(args=[{"xaxis.type": "linear", "yaxis.type": "linear"}], label="LinearXY", method="relayout"),
                                                             dict(args=[{"xaxis.type": "log", "yaxis.type": "log"}],       label="LogXY",    method="relayout"),
                                                             dict(args=[{"xaxis.type": "linear", "yaxis.type": "log"}],    label="LogY",     method="relayout"),
                                                             dict(args=[{"xaxis.type": "log", "yaxis.type": "linear"}],    label="LogX",     method="relayout") ]),
                          direction="down", pad={"r": 10, "t": 10}, showactive=True, x=-0.1, xanchor="left", y=1.5, yanchor="top" ) ] )
    fig.update_layout( template="presentation", title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title, barmode=barmode,
                          font=dict(size=fontsize,family="serif"), legend_title_text='', legend = legend_opt[legend_pos[0]],legend_orientation=legend_pos[1], showlegend=True, paper_bgcolor=margin["color"])
    fig.update_xaxes(matches=matches[0],showline=True,mirror="ticks",zeroline=False,showgrid=True,minor_ticks="inside",tickformat=tickformat[0],range=ranges[0])
    fig.update_yaxes(matches=matches[1],showline=True,mirror="ticks",zeroline=False,showgrid=True,minor_ticks="inside",tickformat=tickformat[1],range=ranges[1])

    if figsize != None: fig.update_layout(width=figsize[0],height=figsize[1])
    if margin["auto"] == False: fig.update_layout(margin=dict(l=margin["margin"][0], r=margin["margin"][1], t=margin["margin"][2], b=margin["margin"][3]))

    return fig

def get_continuous_color(colorscale, intermed):
    """
    Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
    color for any value in that range.

    Plotly doesn't make the colorscales directly accessible in a common format.
    Some are ready to use:

        colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

    Others are just swatches that need to be constructed into a colorscale:

        viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
        colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

    :param colorscale: A plotly continuous colorscale defined with RGB string colors.
    :param intermed: value in the range [0, 1]
    :return: color in rgb string format
    :rtype: str
    """
    if len(colorscale) < 1:
        raise ValueError("colorscale must have at least one color")

    if intermed <= 0 or len(colorscale) == 1:
        return colorscale[0][1]
    if intermed >= 1:
        return colorscale[-1][1]

    for cutoff, color in colorscale:
        if intermed > cutoff:
            low_cutoff, low_color = cutoff, color
        if intermed <= cutoff:
            high_cutoff, high_color = cutoff, color
            break

    # noinspection PyUnboundLocalVariable
    return plotly.colors.find_intermediate_color(
        lowcolor=low_color,
        highcolor=high_color,
        intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
        colortype="rgb",
    )

# OLD FUNCTIONS (TO BE REMOVED)
# def plot_segs(segs, sim_version="minirun4", **kwargs):
#         def to_list(axis):
#             if sim_version == "minirun4":
#                 nice_array = np.column_stack(
#                     [segs[f"{axis}_start"], segs[f"{axis}_end"], np.full(len(segs), None)]
#                 ).flatten()
#             if sim_version == "minirun3":
#                 nice_array = np.column_stack(
#                     [
#                         segs[f"{axis}_start"] * 10,
#                         segs[f"{axis}_end"] * 10,
#                         np.full(len(segs), None),
#                     ]
#                 ).flatten()
#             return nice_array

#         x, y, z = (to_list(axis) for axis in "xyz")

#         trace = go.Scatter3d(x=x, y=y, z=z, **kwargs)

#         return trace  
    
# def draw_tpc(sim_version="minirun4"):
#         anode_xs = np.array([-63.931, -3.069, 3.069, 63.931])
#         anode_ys = np.array([-19.8543, 103.8543])  # two ys
#         anode_zs = np.array([-64.3163, -2.6837, 2.6837, 64.3163])  # four zs
#         if sim_version == "minirun4":  # hit coordinates are in cm
#             detector_center = (0, -268, 1300)
#             anode_ys = anode_ys - (268 + 42)
#             anode_zs = anode_zs + 1300
#         if sim_version == "minirun3":  # hit coordinates are in mm
#             detector_center = (0, 42 * 10, 0)
#             anode_xs = anode_xs * 10
#             anode_ys = anode_ys * 10
#             anode_zs = anode_zs * 10

#         center = go.Scatter3d(
#             x=[detector_center[0]],
#             y=[detector_center[1]],
#             z=[detector_center[2]],
#             marker=dict(size=3, color="green", opacity=0.5),
#             mode="markers",
#             name="tpc center",
#         )
#         anodes = draw_anode_planes(
#             anode_xs, anode_ys, anode_zs, colorscale="ice", showscale=False, opacity=0.1
#         )
#         cathodes = draw_cathode_planes(
#             anode_xs, anode_ys, anode_zs, colorscale="burg", showscale=False, opacity=0.1
#         )
#         return center, anodes, cathodes


# def draw_cathode_planes(x_boundaries, y_boundaries, z_boundaries, **kwargs):
#     traces = []
#     for i_z in range(int(len(z_boundaries) / 2)):
#         for i_x in range(int(len(x_boundaries) / 2)):
#             z, y = np.meshgrid(
#                 np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
#                 np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
#             )
#             x = (
#                 (x_boundaries[i_x * 2] + x_boundaries[i_x * 2 + 1])
#                 * 0.5
#                 * np.ones(z.shape)
#             )
#             traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

#     return traces


# def draw_anode_planes(x_boundaries, y_boundaries, z_boundaries, **kwargs):
#     traces = []
#     for i_z in range(int(len(z_boundaries) / 2)):
#         for i_x in range(int(len(x_boundaries))):
#             z, y = np.meshgrid(
#                 np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
#                 np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
#             )
#             x = x_boundaries[i_x] * np.ones(z.shape)

#             traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

#     return traces


# def draw_light_detectors(data,event_id):
#     try:
#         charge = data["charge/events"][["id", "unix_ts"]][event_id]
#         num_light = data["light/events/data"].shape[0]
#         light = data["light/events", slice(0, num_light)][
#             ["id", "utime_ms"]
#         ]  # we have to try them all, events may not be time ordered
#     except:
#         print("No light information found, not plotting light detectors")
#         return []

#     match_light = match_light_to_charge_event(charge, light, event_id)

#     if match_light is None:
#         print(
#             f"No light event matches found for charge event {event_id}, not plotting light detectors"
#         )
#         return []

#     waveforms_all_detectors = get_waveforms_all_detectors(data, match_light)

#     # make a list of the sum of the waveform and the channel index
#     integral = np.sum(np.sum(waveforms_all_detectors, axis=2), axis=0)
#     max_integral = np.max(integral)
#     index = np.arange(0, waveforms_all_detectors.shape[1], 1)

#     # plot for each of the 96 channels per tpc the sum of the adc values
#     drawn_objects = []
#     drawn_objects.extend(plot_light_traps(data, integral, index, max_integral))

#     return drawn_objects


# def match_light_to_charge_event(charge, light, evid):
#     """
#     Match the light events to the charge event by looking at proximity in time.
#     Use unix time for this, since it should refer to the same time in both readout systems.
#     For now we just take all the light within 1s from the charge event time.
#     """
#     matches = []
#     for i in range(len(light)):
#         if np.abs(light["utime_ms"][i][0] / 1000 - charge["unix_ts"][0]) < 0.5:
#             matches.append([charge["id"][0], light["id"][i]])

#     match_light = []
#     for i in range(len(matches)):
#         if (
#             matches[i][0] == evid
#         ):  # just checking that we get light for the right charge event
#             match_light.append(matches[i][1])
#     if len(match_light) == 0:
#         match_light = None  # no light for this charge event

#     return match_light


# def get_waveforms_all_detectors(data, match_light):
#     """
#     Get the light waveforms for the matched light events.
#     """
#     light_wvfm = data["/light/wvfm", match_light]

#     samples_mod0 = light_wvfm["samples"][:, 0:2, :, :]
#     samples_mod1 = light_wvfm["samples"][:, 2:4, :, :]
#     samples_mod2 = light_wvfm["samples"][:, 4:6, :, :]
#     samples_mod3 = light_wvfm["samples"][:, 6:8, :, :]

#     sipm_channels_module0 = np.array(
#         [2, 3, 4, 5, 6, 7]
#         + [9, 10]
#         + [11, 12]
#         + [13, 14]
#         + [18, 19, 20, 21, 22, 23]
#         + [25, 26]
#         + [27, 28]
#         + [29, 30]
#         + [34, 35, 36, 37, 38, 39]
#         + [41, 42]
#         + [43, 44]
#         + [45, 46]
#         + [50, 51, 52, 53, 54, 55]
#         + [57, 58]
#         + [59, 60]
#         + [61, 62]
#     )

#     sipm_channels_modules = np.array(
#         [4, 5, 6, 7, 8, 9]
#         + [10, 11, 12, 13, 14, 15]
#         + [20, 21, 22, 23, 24, 25]
#         + [26, 27, 28, 29, 30, 31]
#         + [36, 37, 38, 39, 40, 41]
#         + [42, 43, 44, 45, 46, 47]
#         + [52, 53, 54, 55, 56, 57]
#         + [58, 59, 60, 61, 62, 63]
#     )
#     adcs_mod0 = samples_mod0[:, :, sipm_channels_module0, :]
#     adcs_mod1 = samples_mod1[:, :, sipm_channels_modules, :]
#     adcs_mod2 = samples_mod2[:, :, sipm_channels_modules, :]
#     adcs_mod3 = samples_mod3[:, :, sipm_channels_modules, :]

#     all_adcs = np.concatenate((adcs_mod0, adcs_mod1, adcs_mod2, adcs_mod3), axis=1)

#     # instead of a (m, 8, 48, 1000) array, we want a (m, 4, 96, 1000) array
#     # modules instead of tpcs, and 96 channels per module
#     m = len(match_light)
#     all_modules = all_adcs.reshape((m, 4, 96, 1000))

#     # now we make a full array for all the modules
#     # could have been done in one step, but this is easier to read
#     all_detector = all_modules.reshape((m, 384, 1000))

#     return all_detector


# def plot_light_traps(data, n_photons, op_indeces, max_integral):
#     """Plot optical detectors"""
#     drawn_objects = []
#     ys = np.flip(
#         np.array(
#             [
#                 -595.43,
#                 -545.68,
#                 -490.48,
#                 -440.73,
#                 -385.53,
#                 -335.78,
#                 -283.65,
#                 -236.65,
#                 -178.70,
#                 -131.70,
#                 -73.75,
#                 -26.75,
#                 25.38,
#                 75.13,
#                 130.33,
#                 180.08,
#                 235.28,
#                 285.03,
#                 337.15,
#                 384.15,
#                 442.10,
#                 489.10,
#                 547.05,
#                 594.05,
#             ]
#         )
#         / 10
#     )
#     light_width = ys[1] - ys[0]

#     det_bounds = data["/geometry_info/det_bounds/data"]
#     COLORSCALE = plotly.colors.make_colorscale(
#         plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.YlOrRd)[0]
#     )

#     for ix in range(0, det_bounds.shape[0]):
#         for ilight, light_y in enumerate(ys):
#             for iside in range(2):
#                 opid = ilight + iside * len(ys) + ix * len(ys) * 2
#                 if opid not in op_indeces:
#                     continue
#                 xx = np.linspace(det_bounds[ix][0][0][0], det_bounds[ix][0][1][0], 2)
#                 zz = np.linspace(
#                     light_y - light_width / 2 + det_bounds[0][0][1] + 0.25,
#                     light_y + light_width / 2 + det_bounds[0][0][1] - 0.25,
#                     2,
#                 )

#                 xx, zz = np.meshgrid(xx, zz)
#                 light_color = [
#                     [
#                         0.0,
#                         get_continuous_color(
#                             COLORSCALE, intermed=max(0, n_photons[opid]) / max_integral
#                         ),
#                     ],
#                     [
#                         1.0,
#                         get_continuous_color(
#                             COLORSCALE, intermed=max(0, n_photons[opid]) / max_integral
#                         ),
#                     ],
#                 ]

#                 if ix % 2 == 0:
#                     flip = 0
#                 else:
#                     flip = -1

#                 opid_str = f"opid_{opid}"
#                 light_plane = dict(
#                     type="surface", 
#                     y=np.full(xx.shape, det_bounds[ix][0][0][iside + flip])/10 -240,
#                     x=xx/10,
#                     z=zz/10+1300,
#                     opacity=0.4,
#                     hoverinfo="text",
#                     ids=[[opid_str, opid_str], [opid_str, opid_str]],
#                     customdata=[[opid_str, opid_str], [opid_str, opid_str]],
#                     text=f"Optical detector {opid} waveform integral<br>{n_photons[opid]:.2e}",
#                     colorscale=light_color,
#                     showlegend=False,
#                     showscale=False,
#                 )

#                 drawn_objects.append(light_plane)

#     return drawn_objects

# def plot_waveform(data, evid, opid):
#     try:
#         charge = data["charge/events", self.file["id"]][["id", "unix_ts"]]
#         num_light = data["light/events/data"].shape[0]
#         light = data["light/events", slice(0, num_light)][
#             ["id", "utime_ms"]
#         ]  # we have to try them all, events may not be time ordered
#     except:
#         print("No light information found, not plotting light waveform")
#         return []

#     match_light = match_light_to_charge_event(charge, light, self.file["id"])

#     if match_light is None:
#         print(
#             f"No light event matches found for charge event {self.file['id']}, not plotting light waveform"
#         )
#         return []

#     fig = go.Figure()
#     waveforms_all_detectors = get_waveforms_all_detectors(data, match_light)
#     wvfm_opid = waveforms_all_detectors[:, opid, :]
    
#     x = np.arange(0, 1000, 1)
#     y = np.sum(wvfm_opid, axis=0)
#     drawn_objects = go.Scatter(x=x, y=y)
#     fig.add_traces(drawn_objects)
    
#     fig.update_xaxes(title_text='Time [ticks] (1 ns)')
#     fig.update_yaxes(title_text='ADC counts')
#     fig.update_layout(title_text=f'Waveform for optical detector {opid}')
#     return fig



# def get_continuous_color(colorscale, intermed):
#     """
#     Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
#     color for any value in that range.

#     Plotly doesn't make the colorscales directly accessible in a common format.
#     Some are ready to use:

#         colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

#     Others are just swatches that need to be constructed into a colorscale:

#         viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
#         colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

#     :param colorscale: A plotly continuous colorscale defined with RGB string colors.
#     :param intermed: value in the range [0, 1]
#     :return: color in rgb string format
#     :rtype: str
#     """
#     if len(colorscale) < 1:
#         raise ValueError("colorscale must have at least one color")

#     if intermed <= 0 or len(colorscale) == 1:
#         return colorscale[0][1]
#     if intermed >= 1:
#         return colorscale[-1][1]

#     for cutoff, color in colorscale:
#         if intermed > cutoff:
#             low_cutoff, low_color = cutoff, color
#         if intermed <= cutoff:
#             high_cutoff, high_color = cutoff, color
#             break

#     # noinspection PyUnboundLocalVariable
#     return plotly.colors.find_intermediate_color(
#         lowcolor=low_color,
#         highcolor=high_color,
#         intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
#         colortype="rgb",
#     )


"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class ParentPlugin(Plugin):
    """
    A plugin for constructing an array of the parent particles
    pdg code corresponding to the index of the original particle
    in the trajectories array.
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(ParentPlugin, self).__init__(config)

        self.input_products = None
        self.output_products = 'parent_pdg_id'

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        parent_pdg_id = [-1 for ii in range(len(trajectories))]

        traj_ids = trajectories['traj_id']
        vertex_ids = trajectories['vertex_id']
        parent_ids = trajectories['parent_id']
        pdg_ids = trajectories['pdg_id']

        for ii, (traj_id, vertex_id, parent_id) in enumerate(zip(
            traj_ids,
            vertex_ids,
            parent_ids
        )):
            if parent_id != -1:
                parent_index = np.where(
                    (traj_ids == parent_id) & (vertex_ids == vertex_id)
                )[0]
                parent_pdg_id[ii] = pdg_ids[parent_index][0]
        event_products['parent_pdg_id'] = np.array(parent_pdg_id)



import plotly.graph_objects as go
from dash import dcc,html
import numpy as np
import pandas as pd # remove this dependency later
import plotly
import h5py
import traceback

from arrakis_nd.utils.display.utils import custom_plotly_layout, get_continuous_color

class ChargeLightDisplay:
    '''
    This class is used to visualize the events in three different plots:

    '''
    def __init__(
        self,
        geometry_info: dict = {}
    ):
        self.interactions = None
        self.segments = None
        self.stack = None
        self.trajectories = None
        self.charge = None
        self.geometry_info = {}
        self.set_geometry_info(geometry_info)
        self.generate_layout()

    def set_geometry_info(
        self,
        geometry_info,
    ):
        if geometry_info == {}:
            pass
        else:
            try:
                self.geometry_info = geometry_info
                #self.light_traps = self.construct_light_detectors()
                #self.tpc.add_traces(self.light_traps)
            except Exception:
                print(traceback.format_exc())
                #print(f"issue with light traps {e}")


    def construct_detector(self):
        self.tpc = self.construct_tpcs()
        try: aux = self.waveforms.data; print("We already have waveforms!"); del aux
        except AttributeError: self.waveforms = self.construct_waveforms()
        self.larpix = self.construct_larpix()

    def construct_tpcs(self):
        tpc = go.Figure()
        tpc.update_layout(
            scene=dict(
                xaxis_title="x [cm]",
                yaxis_title="y [cm]",
                zaxis_title="z [cm]"
            ),
            title="2x2 TPCs"
        )
        self.tpc_center, self.anodes, self.cathodes = self.draw_tpc()
        tpc.add_traces(self.tpc_center)
        tpc.add_traces(self.anodes)
        tpc.add_traces(self.cathodes)
        return tpc

    def construct_light_detectors(self, waveforms):
        """Plot optical detectors"""
        drawn_objects = []
        
        det_bounds = self.geometry_info["det_bounds"]
        
        channel_map = np.array([0, 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88, 96, 104, 112, 120, 65, 73, 81, 89, 97, 105, 113, 121, 1, 9, 17, 25, 33, 41, 49, 57, 2, 10, 18, 26,
                    34, 42, 50, 58, 66, 74, 82, 90, 98, 106, 114, 122, 67, 75, 83, 91, 99, 107, 115, 123, 3, 11, 19, 27, 35, 43, 51, 59, 4, 12, 20, 28, 36, 44, 52,
                    60, 68, 76, 84, 92, 100, 108, 116, 124, 69, 77, 85, 93, 101, 109, 117, 125, 5, 13, 21, 29, 37, 45, 53, 61, 6, 14, 22, 30, 38, 46, 54, 62, 70,
                    78, 86, 94, 102, 110, 118, 126, 71, 79, 87, 95, 103, 111, 119, 127, 7, 15, 23, 31, 39, 47, 55, 63]) # this maps detector position to detector number
        # we need to invert the mapping because I'm stupid
        channel_map = np.argsort(channel_map)
        channel_map_deluxe = pd.read_csv('arrakis_nd/utils/display/sipm_channel_map.csv', header=0)
        
        xs = []
        ys = []
        zs = []
        for i in range(len(det_bounds)):
            if det_bounds[i][1] == True:
                xs.append([det_bounds[i][0][0][0], det_bounds[i][0][1][0]])
                ys.append([det_bounds[i][0][0][1], det_bounds[i][0][1][1]])
                zs.append([det_bounds[i][0][0][2], det_bounds[i][0][1][2]])
        
        COLORSCALE = plotly.colors.make_colorscale(
            plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.YlOrRd)[0]
        )

        photon_sums = []
        for i in range(len(xs)):
            opid = channel_map[i]
            # get all adc, channel belonging to opid=det_id. We have a numpy array channel_map_deluxe with dtype (det_id, tpc, side, sipm_pos, adc, channel)
            sipms = channel_map_deluxe[channel_map_deluxe['det_id']==opid][['adc', 'channel']].values
            sum_photons = 0
            for adc, channel in sipms:
                wvfm = waveforms[:, int(adc), int(channel), :]
                sum_wvfm = np.sum(wvfm, axis=0) # sum over the events
                sum_photons += np.sum(sum_wvfm, axis=0) # sum over the time
            photon_sums.append(sum_photons)
        max_integral = np.max(photon_sums)
        
        for i in range(len(xs)):
            opid = channel_map[i]
            # get all adc, channel belonging to opid=det_id. We have a numpy array channel_map_deluxe with dtype (det_id, tpc, side, sipm_pos, adc, channel)
            sipms = channel_map_deluxe[channel_map_deluxe['det_id']==opid][['adc', 'channel']].values
            sum_photons = 0
            for adc, channel in sipms:
                wvfm = waveforms[:, int(adc), int(channel), :]
                sum_wvfm = np.sum(wvfm, axis=0)
                sum_photons += np.sum(sum_wvfm, axis=0)
            opid_str = f"opid_{opid}"
            light_color = [
                [
                    0.0,
                    get_continuous_color(
                        COLORSCALE, intermed=max(0, sum_photons/max_integral)
                    ),
                ],
                [
                    1.0,
                    get_continuous_color(
                        COLORSCALE, intermed=max(0, sum_photons/max_integral)
                    ),
                ],
            ]
            light_plane = go.Surface(
                x=xs[i],
                y=ys[i],
                z=[zs[i], zs[i]], # why flip y and z?
                colorscale=light_color,
                showscale=False,
                showlegend=False,
                opacity=0.2,
                hoverinfo="text",
                ids=[[opid_str, opid_str], [opid_str, opid_str]],
                text=f"Optical detector {opid} waveform integral<br>{max(0,sum_photons):.2e}",
            )

            drawn_objects.append(light_plane)
        self.tpc.add_traces(drawn_objects)

    def construct_waveforms(self):
        waveforms = go.Figure()
        waveforms = custom_plotly_layout(waveforms,
                                         xaxis_title ="Time [ticks] (1 ns)",
                                         yaxis_title="ADC counts",
                                         title="Waveform for optical detector")
        return waveforms
    
    def plot_waveform(self, opid, waveforms):
        channel_map_deluxe = pd.read_csv('arrakis_nd/utils/display/sipm_channel_map.csv', header=0)
        sipms = channel_map_deluxe[channel_map_deluxe['det_id']==int(opid)][['adc', 'channel']].values
        sum_wvfm = np.array([0]*1000)
        for adc, channel in sipms:
            wvfm = waveforms[:, int(adc), int(channel), :]
            event_sum_wvfm = np.sum(wvfm, axis=0) # sum over the events
            sum_wvfm += event_sum_wvfm # sum over the sipms

        x = np.arange(0, 1000, 1)
        y = sum_wvfm

        drawn_objects = go.Scatter(x=x, y=y, name=f"Channel sum for light trap {opid}", visible=True, showlegend=True)#, legend_orientation="h")
        self.waveforms.add_traces(drawn_objects)
        for adc, channel in sipms:
            wvfm = waveforms[:, int(adc), int(channel), :]
            sum_wvfm = np.sum(wvfm, axis=0)
            self.waveforms.add_traces(go.Scatter(x=x, y=sum_wvfm, visible="legendonly", showlegend=True, name=f"Channel {adc, channel}"))

    def construct_larpix(self):
        larpix = go.Figure()
        larpix = custom_plotly_layout(larpix,
                                      xaxis_title="x [cm]",
                                      yaxis_title="y [cm]",
                                      title="LArPix detector")
        return larpix

    def draw_tpc(
        self,
        sim_version="minirun5"
    ):
        anode_xs = np.array([-63.931, -3.069, 3.069, 63.931])
        anode_ys = np.array([-19.8543, 103.8543])  # two ys
        anode_zs = np.array([-64.3163, -2.6837, 2.6837, 64.3163])  # four zs
        if sim_version == "minirun4":  # hit coordinates are in cm
            detector_center = (0, -268, 1300)
            anode_ys = anode_ys - (268 + 42)
            anode_zs = anode_zs + 1300
        if sim_version == "minirun3":  # hit coordinates are in mm
            detector_center = (0, 42 * 10, 0)
            anode_xs = anode_xs * 10
            anode_ys = anode_ys * 10
            anode_zs = anode_zs * 10
        if sim_version == "minirun5":  # hit coordinates are in cm
            detector_center = (0, 0, 0)
            anode_ys = anode_ys - 42

        center = go.Scatter3d(
            x=[detector_center[0]],
            y=[detector_center[1]],
            z=[detector_center[2]],
            marker=dict(size=3, color="green", opacity=0.5),
            mode="markers",
            name="tpc center",
        )
        anodes = self.draw_anode_planes(
            anode_xs, anode_ys, anode_zs, colorscale="ice", showscale=False, opacity=0.1
        )
        cathodes = self.draw_cathode_planes(
            anode_xs, anode_ys, anode_zs, colorscale="burg", showscale=False, opacity=0.1
        )
        return center, anodes, cathodes

    def draw_cathode_planes(
        self,
        x_boundaries,
        y_boundaries,
        z_boundaries,
        **kwargs
    ):
        traces = []
        for i_z in range(int(len(z_boundaries) / 2)):
            for i_x in range(int(len(x_boundaries) / 2)):
                z, y = np.meshgrid(
                    np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
                    np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
                )
                x = (
                    (x_boundaries[i_x * 2] + x_boundaries[i_x * 2 + 1])
                    * 0.5
                    * np.ones(z.shape)
                )
                traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

        return traces

    def draw_anode_planes(
        self,
        x_boundaries,
        y_boundaries,
        z_boundaries,
        **kwargs
    ):
        traces = []
        for i_z in range(int(len(z_boundaries) / 2)):
            for i_x in range(int(len(x_boundaries))):
                z, y = np.meshgrid(
                    np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
                    np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
                )
                x = x_boundaries[i_x] * np.ones(z.shape)

                traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

        return traces

    def update_event(
        self,
        interactions,
        segments,
        stack,
        trajectories,
        charge,
    ):
        self.interactions = interactions
        self.segments = segments
        self.stack = stack
        self.trajectories = trajectories
        self.charge = charge
    
    def plot_event(self):
        if self.charge is not None:
            name_to_remove = "calib_final_hits"
            # Create a new list of traces that excludes the ones with the specified name
            new_traces = [trace for trace in self.tpc.data if trace.name != name_to_remove]
            # Update the figure with the new list of traces
            self.tpc.data = new_traces

            charge_hits_traces = go.Scatter3d(
                x=self.charge['x'],
                y=self.charge['y'],
                z=self.charge['z'],
                marker_color=self.charge['Q'],
                marker={
                    "size": 1.75,
                    "opacity": 0.7,
                    "colorscale": "cividis",
                    "colorbar": {
                        "title": "Hit charge [e-]",
                        "titlefont": {"size": 12},
                        "tickfont": {"size": 10},
                        "thickness": 15,
                        "len": 0.5,
                        "xanchor": "left",
                        "x": 0,
                    },
                },
                name="calib_final_hits",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                hovertemplate="<b>x:%{x:.3f}</b><br>y:%{y:.3f}<br>z:%{z:.3f}<br>E:%{customdata:.3f}",
            )

            self.tpc.add_traces(charge_hits_traces)


    def generate_layout(self):
        self.construct_detector()
        self.plot_event()
        self.layout = html.Div(
            [
                dcc.Location(id="url"),
                dcc.Store(id="filename", storage_type="local", data=None),
                dcc.Store(id='data-length', data=0),
                html.H1(children="2x2 event display", style={"textAlign": "center"}),
                html.Div([
                    # This div contains the plot on the left with 50% width of the viewport width
                    html.Div(dcc.Graph(
                        id='charge_light_tpc_plot',
                        style={'height': '80vh', 'width': '45vw'},  # Adjust size as needed
                        figure=self.tpc
                    ), style={'width': '45vw'}),  # This ensures the left plot takes up half the viewport width

                    # This div is the container for the two plots on the right
                    html.Div([
                        # This div contains the top right plot with 35% viewport width and 50% viewport height
                        html.Div(dcc.Graph(
                            id="light-waveform",
                            style={'height': '50vh', 'width': '35vw'},  # Adjust size as needed
                            figure=self.waveforms
                        )),
                        # This div contains the bottom right plot with 35% viewport width and 30% viewport height
                        html.Div(dcc.Graph(
                            id="another-graph",
                            style={'height': '30vh', 'width': '35vw'},  # Adjust size as needed
                            figure=self.larpix
                        )),
                    ], style={'display': 'flex', 'flexDirection': 'column', 'width': '45vw'}),
                ], style={'display': 'flex'})
            ]
        )


"""
"""
import h5py


class Plugin:
    """
    A base class for plugins to Arrakis.

    Plugins are classes which process flow data and produce
    either intermediate data products or arrakis output.
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        All plugins must run super(CustomPlugin, self).__init__(config)
        and specify any required data_products and any output data_products
        so that Arrakis knows they are called in the correct sequence.

        Args:
            config (dict, optional): _description_. Defaults to {}.
        """
        if config is None:
            config = {}
        self.config = config

        """These should be set by the creator of the custom plugin"""
        self.input_products = None
        self.output_products = None

    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        Event processing function for a plugin. The process_event
        function requires both a flow file and arrakis file, as
        well as the dictionary of event_indices corresponding to
        both.  The event_products is a shared dictionary which
        can be updated with intermmediate data used by other
        plugins.


        This function must be overriden by the inherited class.

        NB! Note that any changes one wishes to make to the arrakis_file happen
            in a subtle way.  For example, if you index any dataset from the arrakis_file,
            that will create a new numpy array and will NOT give you a reference
            to the original dataset.  Therefore, any changes you make must be sent back to
            the original arrakis_file in the end!  See the particle examples in the plugin
            folder.

        Args:
            flow_file (h5py.File): _description_
            arrakis_file (h5py.File): _description_
            event_indices (dict): _description_
            event_products (dict): _description_
        """
        raise NotImplementedError("Plugin classes must implement this function!")



import os
import yaml
from dash import (
    Dash,
    dcc,
    html,
    callback_context
)
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import numpy as np
import h5py
import glob

from arrakis_nd.utils.display.set_server import SetServer
from arrakis_nd.utils.display.charge_light_display import ChargeLightDisplay
from arrakis_nd.utils.display.vis_event import VisEvent


class ArrakisDisplay:
    """
    A class to create a Dash app for displaying the results of the ArrakisND analysis.
    """
    def __init__(
        self,
    ):
        """
        Initialize the ArrakisDisplay class.

        Args:
            service_prefix (str, optional): _description_. Defaults to os.getenv("JUPYTERHUB_SERVICE_PREFIX", "/").
            server_url (_type_, optional): _description_. Defaults to "https://jupyter.nersc.gov".
            port (int, optional): _description_. Defaults to 8050.
        """

        server = SetServer()
        config = server.get_config()
        self.service_prefix = config['service_prefix']
        self.server_url = config['server_url']
        self.port = config['port']
        self.jupyter_mode = config['jupyter_mode']

        self.flow_folder = ''
        self.arrakis_folder = ''
        self.flow_files = []
        self.flow_file = ''
        self.arrakis_file = ''
        self.arrakis_files = []
        self.available_events = []
        self.unique_events = []

        self.standard_flow_folders = [
            {'label': 'MiniRun4', 'value':
                '/global/cfs/cdirs/dune/www/data/2x2/simulation/productions/MiniRun4_1E19_RHC/MiniRun4_1E19_RHC.flow/FLOW'},
            {'label': 'MiniRun4.5', 'value':
                '/global/cfs/cdirs/dune/www/data/2x2/simulation/productions/MiniRun4.5_1E19_RHC/inputs_beta3/MiniRun4.5_1E19_RHC.flow/FLOW/0000000'},
            {'label': 'MiniRun5', 'value':
                '/global/cfs/cdirs/dune/www/data/2x2/simulation/productions/MiniRun5_1E19_RHC/MiniRun5_1E19_RHC.flow.beta1/FLOW/0000000'},
        ]

        self.geometry_info = {
            'anode_drift_coordinate': [],
            'det_bounds': [],
            'det_id': [],
            'det_rel_pos': [],
            'drift_dir': [],
            'pixel_coordinates_2D': [],
            'sipm_abs_pos': [],
            'sipm_rel_pos': [],
            'tile_id': [],
        }

        self.interactions = None
        self.segments = None
        self.stack = None
        self.trajectories = None
        self.charge = None
        self.light = None

        self.charge_light_display = ChargeLightDisplay()

        self.construct_app()
        self.construct_widgets()
        self.run_app()

    def adjust_iframe_height(self, height=1000):
        """
        Generates a script to adjust the iframe height for the Dash app when running in Jupyter.
        Parameters:
        height (int): The desired height of the iframe in pixels.
        """
        from IPython.display import display, HTML
        script = f"""
        <script>
        // You might need to adjust the selector depending on your Jupyter environment
        const iframes = document.querySelectorAll('iframe');
        iframes.forEach(function(iframe) {{
            iframe.style.height = '{height}px';
        }});
        </script>
        """
        display(HTML(script))

    def construct_app(self):
        """
        Construct the Dash app.
        """
        if self.jupyter_mode == "inline":
            self.app = Dash(
                __name__,
                requests_pathname_prefix=f"{self.service_prefix}proxy/{self.port}/",
                external_stylesheets=[dbc.themes.SPACELAB]
            )
        else:
            self.app = Dash(
                __name__,
                external_stylesheets=[dbc.themes.SPACELAB]
            )

        """Get the custom style file"""
        with open('arrakis_nd/utils/display/assets/styles.yaml', 'r') as file:
            styles = yaml.safe_load(file)

        """Define the navbar with a dropdown"""
        self.navbar = html.Div(
            children=[
                html.A(
                    href="https://github.com/Neutron-Calibration-in-DUNE/ArrakisND",
                    target="_blank",  # Opens the link in a new tab
                    children=[
                        html.Img(src='/assets/github-mark.png', style={'height': '35px', 'marginRight': '15px'}),
                    ],
                    style={'display': 'flex', 'alignItems': 'center', 'textDecoration': 'none'},
                ),
                # Right side: Menu items
                dbc.DropdownMenu(
                    [
                        dbc.DropdownMenuItem(
                            "A button", id="dropdown-button", n_clicks=0
                        ),
                        dbc.DropdownMenuItem(
                            "Internal link", href="/docs/components/dropdown_menu"
                        ),
                        dbc.DropdownMenuItem(
                            "External Link", href="https://github.com"
                        ),
                        dbc.DropdownMenuItem(
                            "External relative",
                            href="/docs/components/dropdown_menu",
                            external_link=True,
                        ),
                    ],
                    label="Menu",
                ),
            ],
            style=styles['NAVBAR_STYLE']
        )

        """Define the sidebar"""
        self.sidebar = html.Div(
            [
                # DUNE logo and Arrakis Display text
                html.Div(
                    children=[
                        html.Img(src='/assets/2x2.png', style={'height': '100px', 'marginRight': '15px'}),
                        html.H3("2x2 Display"),
                    ],
                    style={'display': 'flex', 'alignItems': 'center'}
                ),

                # Standard FLOW folder selection dropdown
                html.Hr(style={'border': '3px solid #ffffff', 'height': '0px'}),
                html.P("🔍 FLOW/ARRAKIS Folders & Files "),
                dcc.Dropdown(
                    id='standard_flow_dropdown',
                    options=self.standard_flow_folders,
                    style={'color': "#000000"},
                ),
                html.Hr(style={'border': '3px solid #ffffff', 'height': '0px'}),

                # Text box for writing FLOW and Arrakis folders
                dbc.Label("FLOW folder"),
                dbc.Input(
                    placeholder="Enter the FLOW folder",
                    type="text",
                    id='flow_folder_input',
                    size='sm',
                    value=''
                ),
                dbc.Label("ARRAKIS folder"),
                dbc.Input(
                    placeholder="Enter the ARRAKIS folder",
                    type="text",
                    id='arrakis_folder_input',
                    size='sm',
                    value=''
                ),
                html.H2(),

                # Dropdown which lists available flow and arrakis files
                html.Label('FLOW files'),
                dcc.Dropdown(
                    id='flow_dropdown',
                    searchable=True,
                    placeholder='Select a FLOW file...',
                    style={'color': "#000000"}
                ),
                html.Label('ARRAKIS files'),
                dcc.Dropdown(
                    id='arrakis_dropdown',
                    searchable=True,
                    placeholder='Select an ARRAKIS file...',
                    style={'color': "#000000"}
                ),
                html.H2(),

                # Event selector and previous/next buttons
                html.H2(),
                html.Label('Event (spill)'),
                html.Div([
                    dcc.Dropdown(
                        id='event_dropdown',
                        options=self.available_events,
                        searchable=True,
                        placeholder="Select an event...",
                        style={'color': "#000000", 'width': '60%'}
                    ),
                    html.Button('Previous', id='previous_event', style={'width': '20%'}),
                    html.Button('Next', id='next_event', style={'width': '20%'}),
                ], style={'display': 'flex', 'flexDirection': 'row', 'gap': '10px'}),
                html.H2(),
                html.Div(id='event_output'),

                # Event display selector
                html.H2(),
                html.Hr(),
                html.P(
                    "Choose a Display",
                    className="lead"
                ),
                dcc.Dropdown(
                    id='display_dropdown',
                    options=[
                        'Home',
                        'Charge and Light Detectors',
                        'Arrakis',
                        'Clustering'
                    ],
                    value='Home',
                    style={'color': "#000000"}
                ),

                # Highlighting options
                html.Hr(style={'border': '3px solid #ffffff', 'height': '0px'}),
                html.P("FLOW charge types to plot"),
                dcc.Checklist(
                    ['Segments', 'Calib Prompt Hits', 'Calib Final Hits'],
                    ['Calib Prompt Hits'], inline=True
                ),
                html.Hr(style={'border': '1px solid #ffffff', 'height': '0px'}),
                html.Div([
                    html.Div([
                        html.P("Active TPCs"),
                        dcc.Checklist(
                            ['0', '1', '2', '3', '4', '5', '6', '7'],
                            ['0', '1', '2', '3', '4', '5', '6', '7']
                        )], style={'width': '45vw'}),
                    html.Div([
                        html.P("Detectors"),
                        dcc.Checklist(
                            ['0', '1', '2', '3', '4', '5', '6', '7'],
                            ['0', '1', '2', '3', '4', '5', '6', '7']
                        )], style={'width': '45vw'}),
                ], style={'display': 'flex'}),
            ],
            style=styles['SIDEBAR_STYLE'],
        )

        # Define content for the tabs
        self.content = html.Div(id="page-content", style=styles["CONTENT_STYLE"])

        # Define the layout
        self.app.layout = html.Div(style={'overflow': 'scroll'}, children=[
            dcc.Location(id="url"),
            self.navbar,
            self.sidebar,
            self.content,
        ])

    def construct_widgets(self):
        # Callbacks to update the content based on which tab is selected
        @self.app.callback(
            Output("page-content", "children"),
            [Input("display_dropdown", "value")]
        )
        def render_tab_content(pathname):
            if pathname == "Home":
                return html.P("This is the content of the home page!")
            if pathname == "Charge and Light Detectors":
                print("Charge and Light Detectors")
                # print(self.charge_light_display.layout)
                return self.charge_light_display.layout
            elif pathname == "Arrakis":
                return html.P("This is the content of page 2. Yay!")
            elif pathname == "Clustering":
                return html.P("Oh cool, this is page 3!")
            # If the user tries to reach a different page, return a 404 message
            return html.Div(
                [
                    html.H1("404: Not found", className="text-danger"),
                    html.Hr(),
                    html.P(f"The pathname {pathname} was not recognised..."),
                ],
                className="p-3 bg-light rounded-3",
            )

        @self.app.callback(
            Output('flow_folder_input', 'value'),
            Input('standard_flow_dropdown', 'value')
        )
        def update_flow_folder(
            flow_folder
        ):
            return flow_folder

        # Callback to update dropdown options
        @self.app.callback(
            Output('flow_dropdown', 'options'),
            Input('flow_folder_input', 'value'),
        )
        def update_flow_folder_files(
            flow_folder
        ):
            """Check that flow folder has a '/' at the end"""
            if flow_folder:
                if flow_folder[-1] != '/':
                    flow_folder += '/'
            self.flow_folder = flow_folder

            flow_options = []
            if flow_folder and os.path.isdir(flow_folder):
                self.flow_files = sorted([
                    os.path.basename(input_file) for input_file in glob.glob(
                        f"{flow_folder}*.hdf5", recursive=True
                    )
                    if 'FLOW' in input_file
                ])
                flow_options = [
                    {'label': file, 'value': file}
                    for file in self.flow_files
                ]
                return flow_options
            return []

        # Callback to update dropdown options
        @self.app.callback(
            Output('arrakis_dropdown', 'options'),
            Input('arrakis_folder_input', 'value'),
        )
        def update_arrakis_folder_files(
            arrakis_folder
        ):
            """Check that arrakis folder has a '/' at the end"""
            if arrakis_folder:
                if arrakis_folder[-1] != '/':
                    arrakis_folder += '/'

            self.arrakis_folder = arrakis_folder

            arrakis_options = []
            if arrakis_folder and os.path.isdir(arrakis_folder):
                self.arrakis_files = sorted([
                    os.path.basename(input_file) for input_file in glob.glob(
                        f"{arrakis_folder}*.hdf5", recursive=True
                    )
                    if 'ARRAKIS' in input_file
                ])
                arrakis_options = [
                    {'label': file, 'value': file}
                    for file in self.arrakis_files
                ]
                return arrakis_options
            return []

        @self.app.callback(
            Output('event_dropdown', 'options'),
            [Input('flow_dropdown', 'value'), Input('arrakis_dropdown', 'value')],
        )
        def update_available_events(flow_file, arrakis_file):
            self.available_events = []
            if flow_file is not None:
                try:
                    self.flow_file = flow_file
                    with h5py.File(self.flow_folder + flow_file, "r") as flow_file:
                        trajectories = flow_file['mc_truth/trajectories/data']
                        events = trajectories['event_id']
                        for key in self.geometry_info.keys():
                            try:
                                self.geometry_info[key] = flow_file[f'geometry_info/{key}/data'][:]
                            except Exception:
                                print(f"Issue with getting {key} from geometry_info")
                        self.charge_light_display.set_geometry_info(self.geometry_info)
                        self.unique_events = np.unique(events)
                        self.available_events = [
                            {'label': event, 'value': event}
                            for event in self.unique_events
                        ]
                except Exception:
                    pass
            if arrakis_file is not None:
                try:
                    self.arrakis_file = arrakis_file
                    with h5py.File(self.arrakis_folder + arrakis_file, "r") as arrakis_file:
                        pass
                except Exception:
                    pass
            return self.available_events

        @self.app.callback(
            Output('event_dropdown', 'value'),
            [
                Input('previous_event', 'n_clicks'),
                Input('next_event', 'n_clicks')
            ],
            [State('event_dropdown', 'value')]
        )
        def update_event(previous_clicks, next_clicks, current_value):
            triggered_id = callback_context.triggered[0]['prop_id'].split('.')[0] if callback_context.triggered else ''

            if not self.available_events:
                raise PreventUpdate

            current_index = next(
                (i for i, event in enumerate(self.available_events)
                 if event['value'] == current_value), None
            )

            if current_index is None:
                raise PreventUpdate

            new_index = current_index
            if triggered_id == 'previous_event' and current_index > 0:
                new_index = current_index - 1
            elif triggered_id == 'next_event' and current_index < len(self.available_events) - 1:
                new_index = current_index + 1
            else:
                # If we can't go previous or next, raise PreventUpdate to do nothing
                raise PreventUpdate
            return self.available_events[new_index]['value']

        @self.app.callback(
            [Output('event_output', 'children'),
             Output('charge_light_tpc_plot', 'figure')],
            [Input('event_dropdown', 'value')],
        )
        def load_event(event):
            print_output = ''
            if event is not None:
                if self.flow_file:
                    with h5py.File(self.flow_folder + self.flow_file, "r") as flow_file:
                        interactions_events = flow_file['mc_truth/interactions/data']['event_id']
                        #print(np.where(interactions_events == event)[0])
                        segments_events = flow_file['mc_truth/segments/data']['event_id']
                        stack_events = flow_file['mc_truth/stack/data']['event_id']
                        trajectories_events = flow_file['mc_truth/trajectories/data']['event_id']
                        charge_segments = flow_file['mc_truth/calib_final_hit_backtrack/data']['segment_id'].astype(int)
                        charge_fraction = flow_file['mc_truth/calib_final_hit_backtrack/data']['fraction']
                        charge_fraction_mask = (charge_fraction == 0)
                        charge_segments[charge_fraction_mask] = -1
                        non_zero_charge_segments = [row[row != 0] for row in charge_fraction]
                        max_length = len(max(non_zero_charge_segments, key=len))
                        segments_ids = flow_file['mc_truth/segments/data']['segment_id']
                        self.interactions = flow_file['mc_truth/interactions/data'][
                            np.where(interactions_events == event)[0]
                        ]
                        self.segments = flow_file['mc_truth/segments/data'][
                            np.where(segments_events == event)[0]
                        ]
                        self.stack = flow_file['mc_truth/stack/data'][
                            np.where(stack_events == event)[0]
                        ]
                        self.trajectories = flow_file['mc_truth/trajectories/data'][
                            np.where(trajectories_events == event)[0]
                        ]
                        """For charge data we must backtrack through segments"""
                        hits_to_segments = np.any(
                            np.isin(
                                charge_segments[:, :max_length], segments_ids[(segments_events == event)]
                            ),
                            axis=1,
                        )
                        self.charge = flow_file['charge/calib_final_hits/data'][
                            hits_to_segments
                        ]
                        charge_events = flow_file["charge/events/data"]["id"]

                        self.charge_events = flow_file["charge/events/data"][np.where(interactions_events == event)[0]]
                        
                        """Likewise for light data, we must backtrack through segments"""
                        match_light = flow_file['/light/events/data'][:][ flow_file['/charge/events/ref/light/events/ref'][np.where(interactions_events == event)[0],1] ]["id"]
                        self.light = flow_file["light/events/data"][:]  # we have to try them all, events may not be time ordered

                        waveforms_all_detectors = flow_file["light/wvfm/data"]["samples"][match_light]
                        print(match_light)
                        # we have now the waveforms for all detectors matched in time to the event
                    print_output = f"Event: {event} Loaded!"
                    self.charge_light_display.update_event(
                        self.interactions,
                        self.segments,
                        self.stack,
                        self.trajectories,
                        self.charge,
                    )
                    self.charge_light_display.plot_event()
                    self.charge_light_display.construct_light_detectors(waveforms_all_detectors)
                    self.charge_light_display.waveforms = self.charge_light_display.construct_waveforms()
            return print_output, self.charge_light_display.tpc
        
        @self.app.callback(
            Output('light-waveform', 'figure'),
            [Input('charge_light_tpc_plot', 'figure'),
             Input('event_dropdown', 'value'),
             Input('charge_light_tpc_plot', 'clickData')]
        )
        def update_light_waveform(charge_light_tpc_plot, event, click_data):
            if click_data:
                try:
                    if self.flow_file:
                        with h5py.File(self.flow_folder + self.flow_file, "r") as flow_file:
                            interactions_events = flow_file['mc_truth/interactions/data']['event_id']
                                                      
                            """Likewise for light data, we must backtrack through segments"""
                            match_light = flow_file['/light/events/data'][:][ flow_file['/charge/events/ref/light/events/ref'][np.where(interactions_events == event)[0],1] ]["id"]
                            
                            waveforms_all_detectors = flow_file["light/wvfm/data"]["samples"][match_light]
                            opid = click_data['points'][0]['id'].split('_')[1]
                            self.charge_light_display.plot_waveform(opid, waveforms_all_detectors)

                            print("waveform plotted")
                            print(opid)
                            print(self.charge_light_display.waveforms.data)

                            self.charge_light_display.generate_layout() 
                    return self.charge_light_display.waveforms
                except Exception as e:
                    print(e)
                    print("that is not a light trap, no waveform to plot")

    def run_app(self):
        self.app.run_server(
            jupyter_mode=self.jupyter_mode,
            jupyter_server_url=self.server_url,
            host="localhost",
            port=self.port,
        )
        if self.jupyter_mode == "inline":
            self.adjust_iframe_height(height=1500)



"""
Utilities for fitting track data
"""
import numpy as np
from scipy.interpolate import splprep, splev
import warnings
from scipy.integrate import quad, IntegrationWarning

from arrakis_nd.utils.utils import profiler, integrand


@profiler
def fit_track(
    track_t0s,
    track_xyz
):
    """
    This function ...
    """
    try:
        """Sort charge with respect to t0s"""
        combined = sorted(zip(track_t0s, track_xyz), key=lambda x: x[0])
        sorted_t0, sorted_xyz = zip(*combined)
        sorted_t0 = np.array(sorted_t0)
        sorted_xyz = np.array(sorted_xyz)

        """Try to fit a spline curve to the xyz data"""
        tck, u = splprep(
            [sorted_xyz[:, 0], sorted_xyz[:, 1], sorted_xyz[:, 2]],
            s=0
        )

        """Try to integrate this curve"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", IntegrationWarning)
            try:
                curve_length, _ = quad(integrand, 0, 1, args=(tck,))
            except Exception:
                curve_length = 0

        """Get the derivatives at the beginning and ending points"""
        dxdt_start, dydt_start, dzdt_start = splev(0, tck, der=1)
        dxdt_end, dydt_end, dzdt_end = splev(1, tck, der=1)
        dx_start = np.array([dxdt_start, dydt_start, dzdt_start])
        dx_end = np.array([dxdt_end, dydt_end, dzdt_end])
        dx_start_magnitude = np.linalg.norm(dx_start)
        dx_end_magnitude = np.linalg.norm(dx_end)

        if dx_start_magnitude > 0:
            track_dir = [dxdt_start, dydt_start, dzdt_start] / dx_start_magnitude
        else:
            track_dir = [dxdt_start, dydt_start, dzdt_start]

        if dx_end_magnitude > 0:
            track_enddir = [dxdt_end, dydt_end, dzdt_end] / dx_end_magnitude
        else:
            track_enddir = [dxdt_end, dydt_end, dzdt_end]

        """Determine the integrated columnar density"""
        lar_rho = 1.40
        curve_length_rho = lar_rho * curve_length

        """Fill the values"""
        track_data = {
            'track_dir': track_dir,
            'track_enddir': track_enddir,
            'track_len_gcm2': curve_length_rho,
            'track_len_cm': curve_length
        }
    except Exception:
        """Otherwise, these values are undefined"""
        track_data = {
            'track_dir': [0, 0, 0],
            'track_enddir': [0, 0, 0],
            'track_len_gcm2': 0,
            'track_len_cm': 0
        }

    return track_data



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.utils.shower_utils import fit_shower
from arrakis_nd.arrakis.common import shower_data_type
from arrakis_nd.dataset.common import (
    ProcessType, SubProcessType,
    Shower
)


class ShowerPlugin(Plugin):
    """_summary_

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(ShowerPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
            'ancestor_traj_id_map',
            'ancestor_pdg_id_map',
        ]
        self.output_products = ['shower']

        if "shower_size_threshold" not in self.config:
            self.config["shower_size_threshold"] = 10
        self.shower_size_threshold = self.config["shower_size_threshold"]

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_E = trajectories['E_start']
        trajectories_start_subprocess = trajectories['start_subprocess']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']

        """Collect unique fragments"""
        arrakis_fragments = event_products['fragment']
        arrakis_blips = event_products['blip']
        ancestor_traj_index_map = event_products['ancestor_traj_index_map']
        ancestor_traj_id_map = event_products['ancestor_traj_id_map']

        fragment_ids = np.array([fragment['fragment_id'][0] for fragment in arrakis_fragments])
        fragment_vertex_ids = np.array([fragment['vertex_id'][0] for fragment in arrakis_fragments])
        blip_ids = np.array([blip['blip_id'][0] for blip in arrakis_blips])
        blip_vertex_ids = np.array([blip['vertex_id'][0] for blip in arrakis_blips])

        fragment_ancestors = {}
        ancestor_vertex = {}
        blip_ancestors = {}

        for ii, (fragment_id, vertex_id) in enumerate(zip(
            fragment_ids, fragment_vertex_ids
        )):
            ancestor_traj_index = ancestor_traj_index_map[(fragment_id, vertex_id)]
            ancestor_id = ancestor_traj_id_map[(fragment_id, vertex_id)]
            ancestor_vertex[ancestor_id] = vertex_id
            if ancestor_id not in fragment_ancestors.keys():
                fragment_ancestors[ancestor_id] = [ii]
                blip_ancestors[ancestor_id] = []
            else:
                fragment_ancestors[ancestor_id].append(ii)

        for ii, (blip_id, vertex_id) in enumerate(zip(
            blip_ids, blip_vertex_ids
        )):
            ancestor_traj_index = ancestor_traj_index_map[(blip_id, vertex_id)]
            ancestor_id = ancestor_traj_id_map[(blip_id, vertex_id)]
            if ancestor_id in blip_ancestors.keys():
                blip_ancestors[ancestor_id].append(ii)

        """Now that we've collected unique ancestors, lets make a shower for each"""
        for ii, (ancestor_id, frag_indices) in enumerate(fragment_ancestors.items()):
            ancestor_traj_index = ancestor_traj_index_map[(ancestor_id, ancestor_vertex[ancestor_id])]
            ancestor_pdg = trajectories_pdg_ids[ancestor_traj_index]
            ancestor_xyz_start = trajectories_xyz_start[ancestor_traj_index]
            ancestor_pxyz_start = trajectories_pxyz_start[ancestor_traj_index]
            ancestor_xyz_end = trajectories_xyz_end[ancestor_traj_index]
            ancestor_E = trajectories_E[ancestor_traj_index]

            """Combine all hits together for this shower"""
            fragment_hits = np.concatenate([
                track_id_hit_map[(fragment_ids[frag_index], fragment_vertex_ids[frag_index])]
                for frag_index in frag_indices
            ])
            fragment_segments = np.concatenate([
                track_id_hit_segment_map[(fragment_ids[frag_index], fragment_vertex_ids[frag_index])]
                for frag_index in frag_indices
            ])

            """Get the associated t0 values"""
            fragment_t0s = np.concatenate([
                track_id_hit_t0_map[(fragment_ids[frag_index], fragment_vertex_ids[frag_index])]
                for frag_index in frag_indices
            ])

            """Determine track begin and end points"""
            fragment_charge_xyz = np.array([
                charge_x[fragment_hits],
                charge_y[fragment_hits],
                charge_z[fragment_hits]
            ]).T
            fragment_charge_E = charge_E[np.unique(fragment_hits)]

            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            fragment_start_distances = np.sqrt(
                ((fragment_charge_xyz - ancestor_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(fragment_start_distances)

            fragment_start_index = (
                fragment_hits[closest_start_index],
                fragment_segments[closest_start_index]
            )

            """Set track beginning and ending points"""
            arrakis_charge['shower_begin'][fragment_start_index] = 1

            """Fit shower variables"""
            shower_data = fit_shower(
                fragment_t0s,
                fragment_charge_xyz,
                fragment_charge_xyz[closest_start_index],
                ancestor_pxyz_start
            )

            """First we check to see whether a conversion exists, if not then no shower"""
            num_conversions = 0
            for frag_index in frag_indices:
                traj_index = np.where(
                    (trajectories_traj_ids == arrakis_fragments[frag_index]['fragment_id']) &
                    (trajectories_vertex_ids == arrakis_fragments[frag_index]['vertex_id'])
                )
                if trajectories_start_subprocess[traj_index] == SubProcessType.GammaConversion.value:
                    num_conversions += 1

            """Determine shower type"""
            if (
                (abs(ancestor_pdg) == 11) |
                (abs(ancestor_pdg) == 22)
            ):
                shower_type = Shower.Electromagnetic.value
            elif ancestor_pdg == 111:
                shower_type = Shower.Pion0Decay.value
            elif ancestor_pdg == 211:
                shower_type = Shower.PiPlusDecay.value
            elif ancestor_pdg == -211:
                shower_type = Shower.PiMinusDecay.value
            elif ancestor_pdg == 311:
                shower_type = Shower.Kaon0Decay.value
            elif ancestor_pdg == 130:
                shower_type = Shower.KaonLongDecay.value
            elif ancestor_pdg == 310:
                shower_type = Shower.KaonShortDecay.value
            elif ancestor_pdg == 321:
                shower_type = Shower.KaonPlusDecay.value
            elif ancestor_pdg == -321:
                shower_type = Shower.KaonMinusDecay.value
            elif ancestor_pdg == 421:
                shower_type = Shower.D0Decay.value
            elif ancestor_pdg == 411:
                shower_type = Shower.DPlusDecay.value
            elif ancestor_pdg == -411:
                shower_type = Shower.DMinusDecay.value
            elif ancestor_pdg == 3122:
                shower_type = Shower.LambdaDecay.value
            elif ancestor_pdg == 3212:
                shower_type = Shower.Sigma0Decay.value
            elif ancestor_pdg == 3222:
                shower_type = Shower.SigmaPlusDecay.value
            elif ancestor_pdg == 3112:
                shower_type = Shower.SigmaMinusDecay.value
            else:
                shower_type = Shower.Undefined.value

            """Determine start and end of shower"""
            """
            Create the shower object
                shower_data_type = np.dtype([
                    ('event_id', 'i4'),
                    ('shower_id', 'i4'),
                    ('vertex_id', 'i8'),
                    ('blip_ids', 'i4', (1, 400)),
                    ('shower_type', 'i4'),
                    ('start', 'f4', (1, 3)),
                    ('end', 'f4', (1, 3)),
                    ('start_hit', 'f4', (1, 3)),
                    ('end_hit', 'f4', (1, 3)),
                    ('dir', 'f4', (1, 3)),
                    ('enddir', 'f4', (1, 3)),
                    ('dir_hit', 'f4', (1, 3)),
                    ('enddir_hit', 'f4', (1, 3)),
                    ('Evis', 'f4'),
                    ('qual', 'f4'),
                    ('len_gcm2', 'f4'),
                    ('len_cm', 'f4'),
                    ('width', 'f4'),
                    ('angle', 'f4'),
                    ('E', 'f4'),
                    ('truth', 'i4', (1, 20)),
                    ('truthOverlap', 'f4', (1, 20)),
                ])
            """

            shower_data = np.array([(
                event,
                ancestor_id,
                vertex_id,
                shower_type,
                ancestor_xyz_start,
                shower_data['shower_end'],
                [fragment_charge_xyz[closest_start_index]],
                [0, 0, 0],
                ancestor_pxyz_start,
                [0, 0, 0],
                shower_data['shower_dir'],
                [0, 0, 0],
                sum(fragment_charge_E),
                1,
                shower_data['len_gcm2'],
                shower_data['len_cm'],
                shower_data['width'],
                shower_data['angle'],
                ancestor_E,
                [[ancestor_traj_index]+[0]*19],
                [[1]+[0]*19])],
                dtype=shower_data_type
            )

            """Add the new shower to the CAF objects"""
            event_products['shower'].append(shower_data)



import os, socket

class SetServer:
    """
    Set up the Dash app to run in Jupyter or as a standalone server.
    """

    def __init__(self):
        """
        Initialize the SetServer class.
        """
        self.jupyter_mode = self.check_environment()

    def check_environment(self):
        """
        Check if the script is running in a Jupyter notebook or as a standalone script.
        """
        if 'ipykernel' in os.sys.modules:
            from IPython import get_ipython
            if get_ipython() is not None: 
                return "inline"
        return "external"

    def get_config(self):
        """
        Get the appropriate configuration for a Dash application based on the environment.
        """
        if self.jupyter_mode == "inline":
            # Return the configuration for running in a Jupyter notebook
            return {"service_prefix":os.getenv("JUPYTERHUB_SERVICE_PREFIX", "/"),
                    "server_url": "https://jupyter.nersc.gov", 
                    "port": self.find_free_port(),
                    "jupyter_mode": "inline"}
        else:
            # Return the configuration for running as a standalone script
            return {"service_prefix": "/arrakis_display/",
                    "server_url": "http://localhost",
                    "port": self.find_free_port(),
                    "jupyter_mode": None}

    def find_free_port(self, start_port=8050):
        """
        Find a free port to run the Dash app on.
        """
        port = start_port
        while True:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.bind(("localhost", port))
                sock.close()
                return port
            except OSError:
                port += 1
                if port - start_port > 1000:  # Just to avoid an infinite loop
                    raise OSError("No free ports available within 1000 ports of start_port")



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class SegmentPlugin(Plugin):
    """
    A plugin for ...
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        This plugin generates ...

        """
        super(SegmentPlugin, self).__init__(config)

        self.input_products = [
            'parent_pdg_id'
        ]
        self.output_products = None

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        Here we associate (traj_id, vertex_id) pairs to indices in the event_indices['charge']
        mask so that later plugins can easily grab the hits associated to each particle.
        To do this we must first associate (traj_id, vertex_id) to (segment_id) and then
        use the back tracking information in 'mc_truth/calib_final_hit_backtrack/data'
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        segments = flow_file['mc_truth/segments/data'][event_indices['segments']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        charge_back_track = flow_file['mc_truth/calib_final_hit_backtrack/data'][event_indices['charge']]

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_start_process = trajectories['start_process']
        trajectories_start_subprocess = trajectories['start_subprocess']
        segments_traj_ids = segments['traj_id']
        segments_vertex_ids = segments['vertex_id']
        segments_segment_ids = segments['segment_id']
        segments_x = segments['x']
        segments_y = segments['y']
        segments_z = segments['z']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_segment_ids = charge_back_track['segment_id'].astype(int)
        charge_segment_fraction = charge_back_track['fraction']
        charge_segment_fraction_mask = (charge_segment_fraction == 0)
        charge_segment_ids[charge_segment_fraction_mask] = -1

        """Loop over segment ids"""
        for ii, segment_ids in enumerate(charge_segment_ids):
            """Get segment_ids of segments of each hit"""
            segment_ids = segment_ids[(segment_ids != -1)]
            """Determine where these ids are in the segments array"""
            segment_indices = np.array([
                np.where(segments_segment_ids == segment_id)[0][0]
                if segment_id in segments_segment_ids else -1
                for segment_id in segment_ids
            ])
            """Associate this hit to the (traj_id, vertex_id) pairs"""
            for jj, segment_index in enumerate(segment_indices):
                traj_id = segments_traj_ids[segment_index]
                vertex_id = segments_vertex_ids[segment_index]
                traj_index = np.where(
                    (trajectories_traj_ids == traj_id) &
                    (trajectories_vertex_ids == vertex_id)
                )
                arrakis_charge['particle'][ii][jj] = trajectories_pdg_ids[traj_index]
                arrakis_charge['segment_parent'][ii][jj] = event_products['parent_pdg_id'][traj_index]
                arrakis_charge['segment_start_process'][ii][jj] = trajectories_start_process[traj_index]
                arrakis_charge['segment_start_subprocess'][ii][jj] = trajectories_start_subprocess[traj_index]
                arrakis_charge['segment_distance'][ii][jj] = np.sqrt(
                    (charge_x[ii] - segments_x[segment_index]) ** 2 +
                    (charge_y[ii] - segments_y[segment_index]) ** 2 +
                    (charge_z[ii] - segments_z[segment_index]) ** 2
                )
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.utils.utils import fiducialized_vertex
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.dataset.common import Neutrino


class NeutrinoPlugin(Plugin):
    """_summary_

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(NeutrinoPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = []

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        interactions = flow_file['mc_truth/interactions/data'][event_indices['interactions']]
        """Determine neutrino type"""
        argon_neutrino_targets = (interactions['target'] == 18)
        if not sum(argon_neutrino_targets):
            neutrino_type = Neutrino.Undefined.value
            fiducialized = 0
        else:
            """We only want events with one neutrino hitting the argon"""
            if sum(argon_neutrino_targets) > 1:
                return
            neutrino = interactions[argon_neutrino_targets]
            fiducialized = fiducialized_vertex(neutrino['vertex'][0])

            if neutrino['nu_pdg'] == 12:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCElectronNeutrino.value
                else:
                    neutrino_type = Neutrino.CCElectronNeutrino.value
            elif neutrino['nu_pdg'] == -12:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCAntiElectronNeutrino.value
                else:
                    neutrino_type = Neutrino.CCAntiElectronNeutrino.value
            elif neutrino['nu_pdg'] == 14:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCMuonNeutrino.value
                else:
                    neutrino_type = Neutrino.CCMuonNeutrino.value
            elif neutrino['nu_pdg'] == -14:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCAntiMuonNeutrino.value
                else:
                    neutrino_type = Neutrino.CCAntiMuonNeutrino.value
            elif neutrino['nu_pdg'] == 16:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCTauonNeutrino.value
                else:
                    neutrino_type = Neutrino.CCTauonNeutrino.value
            else:
                if neutrino['isCC'] is False:
                    neutrino_type = Neutrino.NCAntiTauonNeutrino.value
                else:
                    neutrino_type = Neutrino.CCAntiTauonNeutrino.value

        neutrino_event_data_type = np.dtype([
            ('event_id', 'i4'),
            ('neutrino_type', 'i4'),
            ('fiducialized', 'i4')
        ])

        neutrino_data = np.array([(
            event,
            neutrino_type,
            fiducialized)],
            dtype=neutrino_event_data_type
        )

        """Add the new track to the CAF objects"""
        event_products['neutrino_event'].append(neutrino_data)
